/*
 NewGameSettingsWindowController.m
 jigplay-macos
 
 Created by François Pessaux on 26/09/10.
 Copyright 2010. All rights reserved.
*/

#import "NewGameSettingsWindowController.h"
#include "misc.h"
#include "alerts_n_errors.h"



@implementation NewGameSettingsWindowController


#pragma mark Start and end of the dialog

- (IBAction) showNewGameSettingsDialog: (id) sender
{
  /* Reset the form and the class internal values to default. */
  [pictureFileNameTextField setStringValue: @""] ;
  [scaleCheckButton setEnabled: NO] ;
  [scaleFactorSlider setFloatValue: 1.0] ;
  [scaleFactorSlider setEnabled: NO] ;
  [widthTextField setStringValue: @"-"] ;
  [heightTextField setStringValue: @"-"] ;
  [averagePieceSizeTextField setStringValue: @"-"] ;
  [resolutionsPopupButton removeAllItems] ;
  [okButton setEnabled: NO] ;
  resolutions = NULL ;
  resolutionsTexts = nil ;

  original_image_width = 0 ;
  original_image_height = 0 ;
  scaled_image_width = 0 ;
  scaled_image_height = 0 ;
  scale_factor = 2.0 ;
  resolutions = NULL ;
  resolutionsTexts = nil ;
  
	[NSApp beginSheet: newGameSettingsDialog
     modalForWindow: mainWindow
      modalDelegate: self
     didEndSelector:
          @selector (didEndNewGameSettingsSheet:returnCode:contextInfo:)
        contextInfo: nil] ;
}



/** Called when the sheet has been closed. Must be implemented otherwise the
    dialog will never close since we set this method as destination selector of
    the sheet.
    By the way, this frees memory used to display the suggested resolutions
    for the jigsaw. */
- (void) didEndNewGameSettingsSheet: (NSWindow *) sheet returnCode: (int) returnCode
                        contextInfo: (void*) contextInfo
{
  /* if (sheet == newGameSettingsDialog) { Do something if needed. } */
  /* Dismiss the sheet. */
  [sheet orderOut: self] ;
  [resolutionsPopupButton removeAllItems] ;
  
  /* Free memory. */
  if (resolutions != NULL) {
    free (resolutions) ;
    resolutions = NULL ;
  }
}



- (IBAction) closeNewGameSettingsDialogCancel: (id) sender
{
  [NSApp endSheet: newGameSettingsDialog] ;
}



- (IBAction) closeNewGameSettingsDialogOk: (id) sender
{
  int width, height, piece_width, piece_height ;
  const char *image_file_name ;
  
  if ([scaleCheckButton state] == NSOnState) {
    width = scaled_image_width ;
    height = scaled_image_height ;
  }
  else {
    width = original_image_width ;
    height = original_image_height ;
    /* To tell the game creation routine not to scale. */
    scale_factor = -1 ;
  }

  image_file_name =
    [[pictureFileNameTextField stringValue]
     cStringUsingEncoding: NSISOLatin1StringEncoding] ;
  
  /* Recover the number of horizontal and vertical pieces from the currently
     selected resolution. */
  int reso_index = [resolutionsPopupButton indexOfSelectedItem] ;
  int nb_horiz_pieces = resolutions[reso_index].nb_horizontally ;
  int nb_vert_pieces = resolutions[reso_index].nb_vertically ;
                    
  /* Compute the possibly cropped dimension due to number of pieces not
     being an exact divisor of the image dimension. */
  piece_width = width / nb_horiz_pieces ;
  piece_height = height / nb_vert_pieces ;
  width = nb_horiz_pieces * piece_width ;
  height = nb_vert_pieces * piece_height ;
  
  /* Extend the width and height so that there is room to move the pieces
     around. Adding 75% should be sufficient. */
  width = width + ((width * 3) / 4) ;
  height = height + ((height * 3) / 4) ;
  
  /* Close the settings dialog and free resolutions stuff by the way. */
  [NSApp endSheet: newGameSettingsDialog] ;
  
  /* Finally, call the PlayAreaView method to start a new game. */
  [playAreaView startNewGame: image_file_name
             withHorizPieces: nb_horiz_pieces
              withVertPieces: nb_vert_pieces
             withScaleFactor: scale_factor
             withWindowWidth: width
            withWindowHeight: height] ;
}



#pragma mark GUI objects reactions

/* Returns TRUE if one resolution is available and selected by default. If none,
   then this return value must trigger the disactivation of the OK button. */
- (bool_t) updateResolutions
{
  int width, height ;
  int i ;
  
  [resolutionsPopupButton removeAllItems] ;
  
  if (resolutions != NULL) {
    free (resolutions) ;
    resolutions = NULL ;
  }
  
  if ([scaleCheckButton state] == NSOnState) {
    width = scaled_image_width ;
    height = scaled_image_height ;
  }
  else {
    width = original_image_width ;
    height = original_image_height ;
  }
  
  resolutions =
    computeRecommendedResolutions (width, height, &nb_found_resolutions) ;
  resolutionsTexts = [NSMutableArray arrayWithCapacity: nb_found_resolutions] ;
  if ((resolutions != NULL) && (resolutionsTexts != nil)) {
    for (i = 0; i < nb_found_resolutions; i++) {
      struct resolution *r = &resolutions[i] ;
      NSString *s =
        [NSString stringWithFormat: @"%d x %d",
          r->nb_horizontally, r->nb_vertically] ;
      [resolutionsTexts insertObject: s atIndex: i] ;
    }     /* End of for (i = 0; i < nb_found_resolutions; i++). */
    
    
  }
  
  [resolutionsPopupButton addItemsWithTitles: (NSArray*) resolutionsTexts] ;
  int index = [resolutionsPopupButton indexOfSelectedItem] ;
  if (index > -1) {
    /* Update the textfield telling the average piece size. */
    [averagePieceSizeTextField
      setIntValue: resolutions[index].average_piece_size] ;
    return (TRUE) ;
  }
  else {
    [averagePieceSizeTextField setStringValue: @"-"] ;
    return (FALSE) ;
  }
}



- (IBAction) resolutionPopupButtonClicked: (id) sender
{
  int index = [resolutionsPopupButton indexOfSelectedItem] ;
  if (index > -1) {
    /* Update the textfield telling the average piece size. */
    [averagePieceSizeTextField
      setIntValue: resolutions[index].average_piece_size] ;
    [okButton setEnabled: TRUE] ;
  }
  else {
    [averagePieceSizeTextField setStringValue: @"-"] ;
    [okButton setEnabled: FALSE] ;
  }  
}



- (void) checkAndUpdateImageFile: (NSString*) fileName
{
  CGImageRef image = NULL ;
  char *suffix ;
  int formatIsJpeg ;
  const char *fName =
    [fileName cStringUsingEncoding: NSISOLatin1StringEncoding] ;
  
  /* Check the format of the image file from its extension. */
  suffix = rindex (fName, (int) '.') ;
  if (suffix == NULL) return ;
  if ((strcasecmp (".jpg", suffix) == 0) ||
      (strcasecmp (".jpeg", suffix) == 0)) {
    /* A priori, format is JPEG. */
    formatIsJpeg = TRUE ;
  }
  else {
    if (strcasecmp (".png", suffix) == 0) {
      /* A priori, format is PNG. */
      formatIsJpeg = FALSE ;
    }
    else {
      /* A priori, non-supported format. */
      return ;
    }
  }
  
  CGDataProviderRef provider = CGDataProviderCreateWithFilename (fName) ;
  if (provider == NULL) {
    [okButton setEnabled: NO] ;
    displayMessage ("Error", "Unable to load image file.", AC_Critical) ;
    return ;
  }
  
  if (formatIsJpeg) {
    image =
      CGImageCreateWithJPEGDataProvider
        (provider, NULL, true, kCGRenderingIntentDefault) ;
  }
  else {
    image =
      CGImageCreateWithPNGDataProvider
        (provider, NULL, true, kCGRenderingIntentDefault) ;
  }
  
  CGDataProviderRelease (provider) ;
  if (image == NULL) {
    [okButton setEnabled: NO] ;
    displayMessage ("Error", "Unable to find an image the file.", AC_Critical) ;
    return ;
  }
  
  /* Extract and update the dimensions of the image. */
  original_image_width = CGImageGetWidth (image) ;
  original_image_height = CGImageGetHeight (image) ;
  
  /* Finally, release the image. */
  CGImageRelease (image) ;
  
  /* And update the file name text field. */
  [pictureFileNameTextField setStringValue: fileName] ;

  /* Update class internal variables. */
  scaled_image_width =
    (int) ((float) original_image_width * scale_factor) ;
  scaled_image_height =
    (int) ((float) original_image_height * scale_factor) ;
  /* Allows scaling. */
  [scaleCheckButton setEnabled: YES] ;
  
  /* Update text fields reporting the image size. */
  if ([scaleCheckButton state] == NSOnState) {
    [scaleFactorSlider setEnabled: YES] ;
    [widthTextField setIntValue: scaled_image_width] ;
    [heightTextField setIntValue: scaled_image_height] ;
  }
  else {
    [scaleFactorSlider setEnabled: NO] ;
    [widthTextField setIntValue: original_image_width] ;
    [heightTextField setIntValue: original_image_height] ;
  }
  
  /* Update the suggested resolutions and finally activate the OK button if at
     least one resolution was found and selected by default. */
  [okButton setEnabled: [self updateResolutions]] ;
}



- (IBAction) selectImageFile: (id) sender
{
  int result, i ;
  NSArray *fileTypes =
    [NSArray arrayWithObjects: @"png", @"jpg", @"jpeg", nil] ;
  NSOpenPanel *panel = [NSOpenPanel openPanel] ;
  [panel setAllowsMultipleSelection: NO] ;
  result =
    [panel runModalForDirectory: NSHomeDirectory ()
                           file: nil types: fileTypes] ;
  
  if (result == NSOKButton) {
    NSArray *filesToOpen = [panel filenames] ;
    i = [filesToOpen count] ;
    
    if (i > 0) {
      NSString *fileNameAsNSString = [filesToOpen objectAtIndex: 0] ;
      [self checkAndUpdateImageFile: fileNameAsNSString] ;
    }   /* End of if (i > 0). */
  }     /* End of if (result == NSOKButton). */
}



- (IBAction) scaleFactorSliderMoved:(id) sender
{
  scale_factor = [scaleFactorSlider floatValue] ;
  
  /* Update class internal variables. */
  scaled_image_width =
    (int) ((float) original_image_width * scale_factor) ;
  scaled_image_height =
    (int) ((float) original_image_height * scale_factor) ;
  /* No need to check if scaling is enabled since when it is not, the slider is
     not activable, hence it can't move, hence this routine can't be called. So
     if we are called, then scaling is allowed and the width and height we must
     display are really the scaled ones. */
  [widthTextField setIntValue: scaled_image_width] ;
  [heightTextField setIntValue: scaled_image_height] ;
  /* Update the suggested resolutions. */
  [okButton setEnabled: [self updateResolutions]] ;
}



- (IBAction) toggleScaleFactorSliderState : (id) sender
{
  [scaleFactorSlider setEnabled: [scaleCheckButton state]] ;
  
  /* Do not set any value in the text fields showing the piece size if there
     is no selected image, i.e. if the image width or height is null. */
  if ((original_image_width <= 0) || (original_image_height <= 0)) return ;
  
  [okButton setEnabled: [self updateResolutions]] ;
}

@end
